#include <stdlib.h>
#include <inttypes.h>
#include <stdio.h>

#define TABLE_SIZE 1024

/*
  Struktur for sidetabel.
  Indeholder blot et array a voidpointers.
*/
typedef struct {

  void* pages[TABLE_SIZE];

} ptabel_t;

/* opretter en sidetabel */
ptabel_t* pt_init() {
  int i;  
  ptabel_t* pagetable;

  pagetable = (ptabel_t*) malloc(sizeof(ptabel_t));  

  /* fyld op med NULL's */
  for(i = 0; i < TABLE_SIZE;i++){
    pagetable->pages[i] = NULL;
  }

  return pagetable;
}

/*
Stter 'page' ind i sidetabellen 'pt'. Den logiske adresse
afgr hvor i sidetabellen 'page' skal placeres. Hvis
'page==NULL' slettes siden fra sidetabellen.
*/
void ptable_set(ptabel_t *pt, void *laddr, void *page) {

  /* vi "overstter" den logiske adresse */
  uintptr_t la = (uintptr_t)laddr;
  uintptr_t d = la & 0xfff;
  uintptr_t p1 = (la >> 22) & 0x3ff;
  uintptr_t p2 = (la >> 12) & 0x3ff;

  /* Vi stter siden ind */
  ptabel_t* pagetable = (ptabel_t*) pt->pages[(unsigned int)p1];

  /* vi skal oprette en ny 'inner' pagetable, hvis der ikke allerede er en */
  if(pagetable == NULL) { 
    pagetable = pt_init();
    pt->pages[(unsigned int)p1] = pagetable;
    printf("Ny undertabel oprettet\n");
  }

  pagetable->pages[(unsigned int)p2] = page;
  if (page == NULL) {
     printf("Slettede side p p1=%u, p2=%u\n", (unsigned int) p1, (unsigned int) p2);
  }
  else {
    printf("Indsat indhold p p1=%u, p2=%u, d=%u!\n", (unsigned int) p1, (unsigned int) p2, (unsigned int) d);
  }
}

/*
  Returnerer pegeren ('page') som er blevet indsat med 'ptable_set'
  for den pgldende logiske adresse 'laddr'. Hvis den logiske
  adresse ikke eksisterer, returneres 'NULL'.
*/
void *ptable_get(ptabel_t *pt, void *laddr) {

  /* vi "overstter" den logiske adresse */
  uintptr_t la = (uintptr_t)laddr;
  uintptr_t d = la & 0xfff;
  uintptr_t p1 = (la >> 22) & 0x3ff;
  uintptr_t p2 = (la >> 12) & 0x3ff;
  
  /* Vi udtager siden */
  ptabel_t* pagetable = pt->pages[(unsigned int)p1];

 /* hvis det er en ugyldig adresse, returneres NULL */
  if(pagetable == NULL) {
    printf("Ugyldig adresse!\n");
    return NULL;
  }
  else {
    /* Som en ekstra finesse ligger vi i vores implementation
       offsetet d til pegeren der er p siden - s kan vi f.eks. under testen
       hive enkelte bogstaver ud af et char-array
    */    
    void* page = (pagetable->pages[(unsigned int)p2])+d;
    if (page == NULL) {
      printf("Ugyldig adresse!\n");
    }
    else {
      printf("Hentede indhold p p1=%u, p2=%u, d=%u!\n", (unsigned int) p1, (unsigned int) p2, (unsigned int) d);
    }
    return page;
  }
}

/* Test program */
int main(int argc, char *argv[]) {

  /* Opret 'ydre' sidetabel */
  ptabel_t* outer = pt_init();

  /* Til brug med page-pegere */
  char test[13] = "Hello World!";
  int i = 42042; // 1010010000111010

  /* til den logiske adresse for der hvor vi vil placere test[] */
  uintptr_t test_int = 14659584; // p1=3, p2=507, d=0
  void* test_la = (void*) test_int;

  /* til den logiske adresse for der hvor vi vil placere i*/
  uintptr_t i_int = 1231757312; // p1=293, p2=690, d=0
  void* i_la = (void*) i_int;

  /* her propper vi ikke noget ind - en ugyldig adresse */
  uintptr_t undef_int = 4242424242; // p1=1011, p2=484, d=434
  void* undef_la = (void*) undef_int;

 
  /* de pegere vi vil proppe ind som 'page'
     Vi bruger pegere til nogle faktiske variabler
     i hukommelsen, sledes at vi kan tjekke at vi
     bde stter korrekt ind og hiver korrekt ud af
     vores sidetabel.
  */
  void* test_page = &test;
  void* i_page = &i;


  /* vi stter ind i vores sidetabel */
  /* ------------------------------- */
  ptable_set(outer, test_la, test_page);
  ptable_set(outer, i_la, i_page);

  /* vi henter fra vores sidetabel */
  /* ----------------------------- */
  char* c_temp;
  int* i_temp;
  
  /* Hent hello world */
  c_temp = (char*) ptable_get(outer, test_la);
  printf("Vi har hentet: %s\n", c_temp);
  
  /* Hent 42042 */
  i_temp = (int*) ptable_get(outer, i_la);
  printf("Vi har hentet: %d\n", *i_temp);
  
  /* Forsg at hente noget p en ugyldig adresse */
  i_temp = ptable_get(outer, undef_la);

  /* Hent W'et ud fra test[] */
  /* Vi bruger offset i ptable_get */
  test_int = 14659590;  // p1=3, p2=507, d=6
  test_la = (void*) test_int;
  c_temp = (char*) ptable_get(outer, test_la);
  printf("Vi har hentet: %c\n", *c_temp);

  /* Vi tester sletning */
  ptable_set(outer, i_la, NULL);
  i_temp = (int*) ptable_get(outer, i_la); // skulle gerne printe 'ugyldig adresse!'

  return 0;
}
